// PIC arithmetic logic unit

// control signals
// asel: 0: f, 1: instruction literal
// bsel: 0: w, 1: bitmask, 2: 'h00, 3: 'h01
// alufn[0] = 0: B, 1: ~B
// alufn[3:1] = select "slow" function
//   0: OR, 1: AND, 2: XOR, 3: <<1, 4: >>1, 5: swap, 6: A, 7: B
// alufn[4] = 0: slow function, 1: A+B+CIN

module alu(f,w,inst,asel,bsel,alufn,cin,result,z,c,dc);
  input [7:0] f,w,inst;
  input [4:0] alufn;
  input cin,asel;
  input [1:0] bsel;
  output [7:0] result;
  output z,c,dc;

  reg [7:0] result;   // outputs
  reg z,c,dc;

  reg [7:0] a,b,xb,slow;   // internal values
  reg [3:0] digit;

  always @ (f or w or inst or asel or bsel or alufn or cin or b or xb or slow or result)
  begin
    c = 1'bx;     // default values for carry condition codes
    dc = 1'bx;
    digit = 4'hx;

    // A operand
    a = asel ? inst[7:0] : f;

    // B operand
    case (bsel)
    2'b00: b = w;
    2'b01: b = 1 << inst[7:5];
    2'b10: b = 8'h00;
    2'b11: b = 8'h01;
    endcase

    // complement B?
    xb = alufn[0] ? ~b : b;

    // "slow" functions
    case (alufn[3:1])
    0: slow = a | xb; 
    1: slow = a & xb;
    2: slow = a ^ xb;
    3: begin slow = {a[6:0],cin}; c = a[7]; end
    4: begin slow = {cin,a[7:1]}; c = a[0]; end
    5: slow = {a[3:0],a[7:4]};
    6: slow = a;
    7: slow = xb;
    endcase

    // select between slow and add/sub
    if (alufn[4] == 0) result = slow;
    else begin
      {c,result} = a + xb + alufn[0];
      {dc,digit} = a[3:0] + xb[3:0] + alufn[0];  // only want dc!
    end

    // compute z condition code
    z = ~|result;
  end
endmodule
