// PIC instruction decode

module decode(inst,skip,
              asel,bsel,alufn,pcsel,skipsel,
	      z_en,c_en,dc_en,w_en,f_en,tris_en);
  input [11:0] inst;
  input skip;
  output asel;		// see alu.v
  output [1:0] bsel;
  output [4:0] alufn;
  output [1:0] pcsel;	// 0: pc+1, 1: goto, 2: call, 3: return
  output [1:0] skipsel;	// 0: no skip, 1: skip, 2: skip if z, 3: skip if ~z
  output z_en;		// 1: update z condition code from alu
  output c_en;		// 1: update c condition code from alu
  output dc_en;		// 1: update dc condition code from alu
  output w_en;		// 1: update w from alu result
  output f_en;		// 1: update f from alu result
  output [2:0] tris_en; // 2,5,6,7: update option,trisa,trisb,trisc

  reg asel;
  reg [1:0] bsel;
  reg [4:0] alufn;
  reg [1:0] pcsel;
  reg [1:0] skipsel;
  reg z_en;
  reg c_en;
  reg dc_en;
  reg w_en;
  reg f_en;
  reg [2:0] tris_en;

  // determine values of control signals for each instruction
  always @ (inst or skip)
  begin
    asel = 1'hx;	// default values (NOP)
    bsel = 2'hx;
    alufn = 5'hxx;
    pcsel = 2'h0;	// pc+1
    skipsel = 2'h0;	// no skip
    z_en = 0;		// z not updated
    c_en = 0;		// c not updated
    dc_en = 0;          // dc not updated
    w_en = 0;           // w not updated
    f_en = 0;		// f not updated
    tris_en = 3'b000;	// option, tris regs not updated
    
    // dispatch on high-order four bits of opcode.  Some instructions
    // will have to look at additional opcode bits.
    if (!skip) case (inst[11:8])
    4'b0000:
      begin
        w_en = !inst[5];
        f_en = inst[5];
        case (inst[7:6])
        2'b00: begin			// NOP, MOVWF
                 bsel = 0;		// B = w;
		 alufn = 5'b01110;	// alufn = B
                 if (!inst[5]) tris_en = inst[2:0];
               end
        2'b01: begin			// CLRF, CLRW
	         bsel = 2;		// B = 0
		 alufn = 5'b01110;	// alufn = B
		 z_en = 1;
	       end
        2'b10: begin			// SUBWF
		 asel = 0;		// A = f
		 bsel = 0;		// B = w
		 alufn = 5'b1xxx1;	// alufn = A-B
		 c_en = 1;
		 dc_en = 1;
		 z_en = 1;
               end
        2'b11: begin			// DECF
		 asel = 0;		// A = f
		 bsel = 3;		// B = 1
		 alufn = 5'b1xxx1;	// alufn = A-B
		 z_en = 1;
	       end
        endcase
      end
    4'b0001:
      begin
        w_en = !inst[5];
        f_en = inst[5];
        case (inst[7:6])
        2'b00: begin			// IORWF
		 asel = 0;		// A = f
		 bsel = 0;		// B = w
		 alufn = 5'b00000;	// alufn = A or B
		 z_en = 1;
               end
        2'b01: begin			// ANDWF
		 asel = 0;		// A = f
		 bsel = 0;		// B = w
		 alufn = 5'b00010;	// alufn = A and B
		 z_en = 1;
               end
        2'b10: begin			// XORWF
		 asel = 0;		// A = f
		 bsel = 0;		// B = w
		 alufn = 5'b00100;	// alufn = A xor B
		 z_en = 1;
               end
        2'b11: begin			// ADDWF
		 asel = 0;		// A = f
		 bsel = 0;		// B = w
		 alufn = 5'b1xxx0;	// alufn = A+B
		 c_en = 1;
		 dc_en = 1;
		 z_en = 1;
               end
        endcase
      end
    4'b0010:
      begin
        w_en = !inst[5];
        f_en = inst[5];
        case (inst[7:6])
        2'b00: begin			// MOVF
		 asel = 0;		// A = f
		 alufn = 5'b0110x;	// alufn = A
		 z_en = 1;
	       end
        2'b01: begin			// COMF
		 asel = 0;		// A = f
		 bsel = 2;		// B = 0
		 alufn = 5'b00101;	// alufn = A xor ~B
		 z_en = 1;
	       end
        2'b10: begin			// INCF
		 asel = 0;		// A = f
		 bsel = 3;		// B = 1
		 alufn = 5'b1xxx0;	// alufn = A + B
		 z_en = 1;
	       end
        2'b11: begin			// DECFSZ
		 asel = 0;		// A = f
		 bsel = 3;		// B = 1
		 alufn = 5'b1xxx1;	// alufn = A - B
		 skipsel = 2;		// skip if z
	       end
        endcase
      end
    4'b0011:
      begin
        w_en = !inst[5];
        f_en = inst[5];
        case (inst[7:6])
        2'b00: begin			// RRF
		 asel = 0;		// A = f
		 alufn = 5'b0100x;	// alufn = A >> 1
		 c_en = 1;
	       end
        2'b01: begin			// RLF
		 asel = 0;		// A = f
		 alufn = 5'b0011x;	// alufn = A << 1
		 c_en = 1;
	       end
        2'b10: begin			// SWAPF
		 asel = 0;		// A = f
		 alufn = 5'b0101x;	// alufn = swap nibbles in A
	       end
        2'b11: begin			// INCFSZ
		 asel = 0;		// A = f
		 bsel = 3;		// B = 1
		 alufn = 5'b1xxx0;	// alufn = A + B
		 skipsel = 2;		// skip if z
	       end
        endcase
      end
    4'b0100:				// BCF
      begin
        asel = 0;			// A = f
	bsel = 3;			// B = 1 << inst[7:5]
	alufn = 5'b00011;		// alufn = A & ~B
	f_en = 1;
      end 
    4'b0101:				// BSF
      begin
        asel = 0;			// A = f
	bsel = 3;			// B = 1 << inst[7:5]
	alufn = 5'b00000;		// alufn = A | B
	f_en = 1;
      end 
    4'b0110:				// BTFSC
      begin
        asel = 0;			// A = f
	bsel = 3;			// B = 1 << inst[7:5]
	alufn = 5'b00010;		// alufn = A & B
	skipsel = 2;			// skip if z
      end 
    4'b0111:				// BTFSS
      begin
        asel = 0;			// A = f
	bsel = 3;			// B = 1 << inst[7:5]
	alufn = 5'b00010;		// alufn = A & B
	skipsel = 3;			// skip if ~z
      end 
    4'b1000:				// RETLW
      begin
        asel = 1;			// A = inst[7:0]
	alufn = 5'b0110x;		// alufn = A
	w_en = 1;
	pcsel = 3;			// pc = return
	skipsel = 1;			// skip
      end 
    4'b1001:				// CALL
      begin
	pcsel = 2;			// pc = call
	skipsel = 1;			// skip
      end 
    4'b1010, 4'b1011:			// GOTO
      begin
	pcsel = 1;			// pc = goto
	skipsel = 1;			// skip
      end 
    4'b1100:				// MOVLW
      begin
        asel = 1;			// A = inst[7:0]
	alufn = 5'b0110x;		// alufn = A
	w_en = 1;
      end 
    4'b1101:				// ORLW
      begin
        asel = 1;			// A = inst[7:0]
	bsel = 0;			// B = w
	alufn = 5'b00000;		// alufn = A or B
	w_en = 1;
      end 
    4'b1110:				// ANDLW
      begin
        asel = 1;			// A = inst[7:0]
	bsel = 0;			// B = w
	alufn = 5'b00010;		// alufn = A and B
	w_en = 1;
      end 
    4'b1111:				// XORLW
      begin
        asel = 1;			// A = inst[7:0]
	bsel = 0;			// B = w
	alufn = 5'b00100;		// alufn = A xor B
	w_en = 1;
      end 
    endcase
  end
endmodule
