// PIC16C5X microcontroller core

// requires asynchronous instruction memory (ROM)
//   pc[PCMSB:0] instruction address
//   idata[11:0] instruction data

module pic16c5x(clk,reset,pc,idata,
		port_a_in,port_a_out,tris_a,
		port_b_in,port_b_out,tris_b,
		port_c_in,port_c_out,tris_c);
  parameter PCMSB = 10;  // 2048 instruction locations max

  input clk;
  input reset;
  output [PCMSB:0] pc;
  input [11:0] idata;
  input [3:0] port_a_in;
  input [7:0] port_b_in;
  input [7:0] port_c_in;
  output [3:0] port_a_out,tris_a;
  output [7:0] port_b_out,tris_b;
  output [7:0] port_c_out,tris_c;

  // PIC registers
  reg [7:0] regfile[63:0];  // register file
  reg [7:0] globalregs[7:0]; 
  reg [PCMSB:0] pc;         // program counter
  reg [PCMSB:0] stack1;     // top of stack register
  reg [PCMSB:0] stack2;     // 2nd stack register
  reg [11:0] inst;          // instruction register
  reg [7:0] w;              // accumulator
  reg [5:0] option;
  reg [7:0] tmr0,prescaler;
  reg [7:0] status;
  reg [6:0] fsr;
  reg [3:0] port_a_out,tris_a;
  reg [7:0] port_b_out,tris_b;
  reg [7:0] port_c_out,tris_c;
  reg skip;

  // prescaler load value
  reg [7:0] prescaler_ld;
  always @ (option)
  begin
    prescaler_ld = 0;
    if (option[3] == 0) case (option[2:0])
    3'b000: prescaler_ld = 1;	// div by 2
    3'b001: prescaler_ld = 3;	// div by 4
    3'b010: prescaler_ld = 7;	// div by 8
    3'b011: prescaler_ld = 15;	// div by 16
    3'b100: prescaler_ld = 31;	// div by 32
    3'b101: prescaler_ld = 63;	// div by 64
    3'b110: prescaler_ld = 127;	// div by 128
    3'b111: prescaler_ld = 255;	// div by 256
    endcase
  end

  // instruction register (if skip load NOP)
  always @ (posedge clk) inst <= idata;

  // There are four register banks, each with 32 8-bit regs.
  // The first 16 regs in each bank are mapped to 16 global regs
  // (8 special purpose regs, 8 general registers), so there are
  // really only 4*16 + 8 = 72 regs in the regfile.  Remap the
  // the address so that the global regs appear at the top half
  // of the regfile address space.
  // Reg addr comes from FSR instead of instruction when indirecting
  // through register 0.
  wire [6:0] reg_addr;
  assign reg_addr = (inst[4:0] == 0) ? {~fsr[4:3],fsr[6:5],fsr[2:0]} :
                                       {~inst[4:3],fsr[6:5],inst[2:0]};

  // regs 0 - 7 are various special purpose regs.  Select the right value...
  reg [7:0] sdata;
  always @ (reg_addr or tmr0 or pc or status or fsr or
            port_a_in or port_b_in or port_c_in)
  case (reg_addr[2:0])
  0: sdata = 8'h00;
  1: sdata = tmr0;
  2: sdata = pc[7:0];
  3: sdata = status;
  4: sdata = {1'b1,fsr};
  5: sdata = {4'b0,port_a_in};
  6: sdata = port_b_in;
  7: sdata = port_c_in;
  endcase

  // select between special purpose regs and regfile data
  reg [7:0] f;
  always @ (reg_addr or sdata /*or globalregs or regfile*/)
    case (reg_addr[6:5])
      default: f = regfile[reg_addr[5:0]];
      2'b10: f = globalregs[reg_addr[2:0]];
      2'b11: f = sdata;
    endcase

  // decode current instruction into appropriate control signals
  wire asel,z_en,c_en,dc_en,w_en,f_en;
  wire [1:0] bsel,pcsel,skipsel;
  wire [2:0] tris_en;
  wire [4:0] alufn;
  decode decode0(inst,skip,
                 asel,bsel,alufn,pcsel,skipsel,
	         z_en,c_en,dc_en,w_en,f_en,tris_en);

  // alu
  wire z,c,dc;
  wire [7:0] result;
  alu alu0(f,w,inst[7:0],asel,bsel,alufn,status[0],result,z,c,dc);

  // final state changes at end of instruction execution
  always @ (posedge clk)
  begin
    if (reset)
      begin
        pc <= -1;
        option <= 6'b111111;
        status[7:3] <= 5'b00011;
        fsr[6:5] <= 2'b11;
	tris_a <= 4'b1111;
	tris_b <= 8'b11111111;
	tris_c <= 8'b11111111;
      end
    else
      begin
        // update program counter
        case (pcsel)
	0: pc <= pc + 1;  // normal
	1: pc <= {status[6:5],inst[8:0]};  // goto
	2: begin  // call
             stack2 <= stack1;
             stack1 <= pc;
             pc <= {status[6:5],1'b0,inst[7:0]};
           end
	3: begin  // return
            pc <= stack1;
            stack1 <= stack2;
           end
	endcase

	case (skipsel)
	0: skip <= 0;
	1: skip <= 1;
	2: skip <= z;
	3: skip <= ~z;
	endcase

	// update condition codes
	if (c_en) status[0] <= c;  
	if (dc_en) status[1] <= dc;  
	if (z_en) status[2] <= z;  

	// update W
        if (w_en) w <= result;

        // update tmr0: do this before possible write to tmr0 (see below)
	// for now clock source is always instruction clock), ie, as if
        // option[5] was always 0.
	prescaler <= (prescaler != 0) ? prescaler - 1 : prescaler_ld;
	if (prescaler == 0) tmr0 <= tmr0 + 1;

	// update register file or global regs
	if (f_en) begin
          if (reg_addr[6]) begin           // write global regs
            if (reg_addr[5])
              case (reg_addr[2:0])
	      1: tmr0 <= result;
	      2: begin pc <= {status[6:5],1'b0,result}; skip <= 1; end
	      3: status <= result;
	      4: fsr <= result[6:0];
	      5: port_a_out <= result[3:0];
	      6: port_b_out <= result;
	      7: port_c_out <= result;
              endcase
	    else globalregs[reg_addr[2:0]] <= result;
          end else regfile[reg_addr[5:0]] <= result;   // write regfile
	end

        // update option, tris regs
        if (tris_en == 3'b010) option <= result[5:0];
        if (tris_en == 3'b101) tris_a <= result[3:0];
        if (tris_en == 3'b110) tris_b <= result;
        if (tris_en == 3'b111) tris_c <= result;

      end
  end

  /* no tristate inside the FPGA !!!
  // i/o ports
  assign port_a[0] = tris_a[0] ? 1'bz : port_a_reg[0];
  assign port_a[1] = tris_a[1] ? 1'bz : port_a_reg[1];
  assign port_a[2] = tris_a[2] ? 1'bz : port_a_reg[2];
  assign port_a[3] = tris_a[3] ? 1'bz : port_a_reg[3];

  assign port_b[0] = tris_b[0] ? 1'bz : port_b_reg[0];
  assign port_b[1] = tris_b[1] ? 1'bz : port_b_reg[1];
  assign port_b[2] = tris_b[2] ? 1'bz : port_b_reg[2];
  assign port_b[3] = tris_b[3] ? 1'bz : port_b_reg[3];
  assign port_b[4] = tris_b[4] ? 1'bz : port_b_reg[4];
  assign port_b[5] = tris_b[5] ? 1'bz : port_b_reg[5];
  assign port_b[6] = tris_b[6] ? 1'bz : port_b_reg[6];
  assign port_b[7] = tris_b[7] ? 1'bz : port_b_reg[7];

  assign port_c[0] = tris_c[0] ? 1'bz : port_c_reg[0];
  assign port_c[1] = tris_c[1] ? 1'bz : port_c_reg[1];
  assign port_c[2] = tris_c[2] ? 1'bz : port_c_reg[2];
  assign port_c[3] = tris_c[3] ? 1'bz : port_c_reg[3];
  assign port_c[4] = tris_c[4] ? 1'bz : port_c_reg[4];
  assign port_c[5] = tris_c[5] ? 1'bz : port_c_reg[5];
  assign port_c[6] = tris_c[6] ? 1'bz : port_c_reg[6];
  assign port_c[7] = tris_c[7] ? 1'bz : port_c_reg[7];
  */

endmodule
